#include "UndefinedTask.hpp"
#include "../types.hpp"

#include <chrono>
#include <cstring>
#include <iostream>
#include <thread>
#include <sys/socket.h>

#include "../../Context.hpp"
#include "../../events/types.hpp"
#include "../../packets/base/GenericPacket.hpp"
#include "../../packets/search/SearchPacketData.hpp"


namespace drp::task {


void UndefinedTask::handle(Context& context) {
    // TODO(Faraphel): If .status is UNDEFINED, look for a server.
    //  if alone, become a server (if can emit).
    //  if everyone is UNKNOWN, elect a server (easiest to join / highest mac address, etc...)
    //  if a server is found, become a client.

    // check if no more peers have been found.
    if (
        std::chrono::high_resolution_clock::now() - context.latestPeerDiscovery >
        std::chrono::milliseconds(5000)
    ) {
        // verify if there are peers
        // TODO(Faraphel): the map is never empty since there is a least the local client
        if (context.remotePeers.empty()) {
            // if we are alone in the network

            // check if we are capable of being a server
            if (!context.me.serverEnabled)
                return;

            // set ourselves as the server
            context.server = nullptr;
            context.me.status = TaskType::SERVER;
            return;
        }

        // look for a server among the peers
        const auto& server = std::ranges::find_if(
            context.remotePeers,
            [&](const auto& peer) { return peer.second->information.status == TaskType::SERVER; }
        );

        if (server != context.remotePeers.end()) {
            // if a server have been found, use it
            context.server = server->second;
            context.me.status = TaskType::CLIENT;
            return;
        }

        // TODO(Faraphel): check if we have the lowest average ping out of all the peers. If yes, become the server.
        //   Others peers will connect to us on their next loop.
        // TODO(Faraphel): use a unique ID instead ?
        if (!context.me.serverEnabled)
            return;

        // check we have the lowest latency out of all the peers
        if (std::all_of(
            context.remotePeers.begin(),
            context.remotePeers.end(),
            [&](const auto& peer) {
                return peer.second->information.latencyAverage > context.me.latencyAverage;
            }
        )) {
            // set ourselves as the server
            context.server = nullptr;
            context.me.status = TaskType::SERVER;
            return;
        }
    }

    // prepare a search message
    packet::GenericPacket packet {};
    packet::GenericPacketContent packetContent {};
    packet::SearchPacketData packetData {};

    // broadcast message
    packet.channel = 0;
    packet.securityMode = static_cast<std::uint8_t>(packet::SecurityMode::PLAIN);
    // search message with the time of the message being sent
    packetContent.eventType = static_cast<std::uint8_t>(event::EventType::SEARCH);
    packetData.timestamp = std::chrono::high_resolution_clock::now();
    std::memcpy(&packetContent.data, &packetData, sizeof(packetData));
    packet.setContent(packetContent);

    // send the search message
    if (sendto(
        context.socket,
        &packet,
        sizeof(packet),
        0,
        context.broadcastAddressInfo->ai_addr,
        context.broadcastAddressInfo->ai_addrlen
    ) == -1)
        std::cerr << "[Sender] Could not send search event: " << strerror(errno) << std::endl;

    std::this_thread::sleep_for(std::chrono::seconds(1));
}


}
